--
-- Calamath's Addon Diagnosis [CDIAG]
--
-- Copyright (c) 2020 Calamath
--
-- This software is released under the Artistic License 2.0
-- https://opensource.org/licenses/Artistic-2.0
--

-- ---------------------------------------------------------------------------------------
-- CT_SimpleAddonFramework: Simple Add-on Framework Template Class              rel.1.0.11
-- ---------------------------------------------------------------------------------------
local CT_SimpleAddonFramework = ZO_Object:Subclass()
function CT_SimpleAddonFramework:New(...)
	local newObject = setmetatable({}, self)
	newObject:Initialize(...)
	newObject:OnInitialized(...)
	return newObject
end
function CT_SimpleAddonFramework:Initialize(name, attributes)
	if type(name) ~= "string" or name == "" then return end
	self._name = name
	self._isInitialized = false
	if type(attributes) == "table" then
		for k, v in pairs(attributes) do
			if self[k] == nil then
				self[k] = v
			end
		end
	end
	self.authority = self.authority or {}
	self._class = {}
	self._shared = nil
	self._external = {
		name = self.name or self._name, 
		version = self.version, 
		author = self.author, 
		RegisterClassObject = function(_, ...) self:RegisterClassObject(...) end, 
	}
	assert(not _G[name], name .. " is already loaded.")
	_G[name] = self._external
	self:ConfigDebug()
	EVENT_MANAGER:RegisterForEvent(self._name, EVENT_ADD_ON_LOADED, function(event, addonName)
		if addonName ~= self._name then return end
		EVENT_MANAGER:UnregisterForEvent(self._name, EVENT_ADD_ON_LOADED)
		self:OnAddOnLoaded(event, addonName)
		self._isInitialized = true
	end)
end
function CT_SimpleAddonFramework:ConfigDebug(arg)
	local debugMode = false
	local key = HashString(GetDisplayName())
	if LibDebugLogger then
		for _, v in pairs(arg or self.authority or {}) do
			if key == v then debugMode = true end
		end
	end
	if debugMode then
		self._logger = self._logger or LibDebugLogger(self._name)
		self.LDL = self._logger
	else
		self.LDL = {
			Verbose = function() end, 
			Debug = function() end, 
			Info = function() end, 
			Warn = function() end, 
			Error = function() end, 
		}
	end
	self._isDebugMode = debugMode
end
function CT_SimpleAddonFramework:RegisterClassObject(className, classObject)
	if className and classObject and not self._class[className] then
		self._class[className] = classObject
		return true
	else
		return false
	end
end
function CT_SimpleAddonFramework:HasAvailableClass(className)
	if className then
		return self._class[className] ~= nil
	end
end
function CT_SimpleAddonFramework:CreateClassObject(className, ...)
	if className and self._class[className] then
		return self._class[className]:New(...)
	end
end
function CT_SimpleAddonFramework:OnInitialized(name, attributes)
--  Available when overridden in an inherited class
end
function CT_SimpleAddonFramework:OnAddOnLoaded(event, addonName)
--  Should be Overridden
end


-- ---------------------------------------------------------------------------------------
-- CAddonDiagnosis
-- ---------------------------------------------------------------------------------------
local CDIAG = CT_SimpleAddonFramework:New("CAddonDiagnosis", {
	name = "CAddonDiagnosis", 
	version = "3.42.0", 
	author = "Calamath", 
	savedVarsDB = "CAddonDiagnosisDB", 
	savedVarsSV = "CAddonDiagnosisSV", 
	savedVarsVersion = 1, 
	authority = {2973583419,210970542}, 
})

local L = GetString
local am = GetAddOnManager()

local NORMAL_COLOR = "dcdcdc"
local DARK_ORANGE = "ffa500"
local YELLOW = "ffff00"
local RED = "ff0000"
local MAGENTA = "ff00ff"
local LIME = "00ff00"
local CORNFLOWER_BLUE = "6495ed"
local ICON_TEXTURE = {
	["ENABLED"] = "|c00ff00|t100%:100%:esoui/art/miscellaneous/check_icon_32.dds:inheritcolor|t|r", 
	["DISABLED"] = "|cff0000|t80%:80%:esoui/art/castbar/forbiddenaction.dds:inheritcolor|t |r", 
	["ERROR"] = "|cff0000|t100%:100%:esoui/art/miscellaneous/new_icon.dds:inheritcolor|t|r", 
	["ERROR_VERSION"] = "|cff00ff|t100%:100%:esoui/art/miscellaneous/new_icon.dds:inheritcolor|t|r", 
	["MISSING"] = "|cff0000|t100%:100%:esoui/art/miscellaneous/help_icon.dds:inheritcolor|t|r", 
	["LOCKED"] = "|cff0000|t100%:100%:esoui/art/miscellaneous/status_locked.dds:inheritcolor|t|r", 
	["OUTDATED"] = "|cffff00|t100%:100%:esoui/art/miscellaneous/eso_icon_warning.dds:inheritcolor|t|r", 
}

-- ---------------------------------------------------------------------------------------
-- Helper Functions
-- ---------------------------------------------------------------------------------------
local function GetTableElements(tbl)
	local n = 0
	if type(tbl) == "table" then
		for _ in pairs(tbl) do
			n = n + 1
		end
	end
	return n
end

local function GetFilename(filePath)
	if filePath then
		return zo_strmatch(filePath, "[^/]+$")
	end
end

local function GetAddonPath(filePath)
	if filePath then
		return zo_strmatch(filePath, "^/?([^/]+)/")
	end
end

local function GetFileExtension(filename)
	if filename then
		filenameWithoutExt, ext = zo_strmatch(filename, "(.+)%.([^%.]+)$")
		if ext == nil then
			filenameWithoutExt = filename
		end
		return ext, filenameWithoutExt
	end
end

local function GetTableKeyForValue(table, value)
	for k, v in pairs(table) do
		if v == value then 
			return k
		end
	end
	return nil
end

local function Decolorize(str)
	if type(str) == "string" then
		return str:gsub("|[cC]%x%x%x%x%x%x", ""):gsub("|[rR]", "")
	else
		return str
	end
end

local function Colorize(colorDef, str, option)
--	colorDef : ZO_ColorDef object or hexadecimal RGB notation (usually 6 characters, ignoring alpha)
--	str      : string
--	option   : string for specifying feature options [optional]
--				  nil      --> same as ZO_ColorDef:Colorize(), regardless of whether it is colored text or not.
--				 "FILL"    --> fills the entire text with the specified color, including colored text
--				 "DEFAULT" --> passes colored text unchanged and fills only standard color text with specified color
	local rgbHex
	local tbl
	if type(colorDef) == "string" then
		rgbHex = colorDef:match("(%x%x%x%x%x%x)")	-- checking hexadecimal RGB notation
	elseif type(colorDef) == "table" then
		if colorDef.ToHex and colorDef.Colorize then	-- checking ZO_ColorDef
			rgbHex = colorDef:ToHex()
		end
	end
	if rgbHex and type(str) == "string" then
		if option == "FILL" then
			tbl = { "|c", rgbHex, Decolorize(str), "|r" }
		elseif option == "DEFAULT" then
			tbl = { "|c", rgbHex, str:gsub("(|[cC]%x%x%x%x%x%x.-|[rR])", "|r%1|c"..rgbHex), "|r" }
		else
			tbl = { "|c", rgbHex, str, "|r" }
		end
		return table.concat(tbl)
	else
		return str
	end
end

local function RemoveTexture(str)
	if type(str) == "string" then
		return str:gsub("|t[^|]+|t", "")
	else
		return str
	end
end

local cached_pcall = pcall
local pcall = function(callback, ...)
	local status, result = cached_pcall(callback, ...)
	if not status then
--		assert(status, "[CDIAG]ERR: " .. result:match("[^%c]*"):match("[^/]+$") or result)
		CDIAG.LDL:Debug("ERR: %s", result:match("[^%c]*"):match("[^/]+$") or result)
	end
	return status, result
end

local cprint, cprintf
if CHAT_ROUTER then
	cprint = function(text)
		pcall(function(text) CHAT_ROUTER:AddSystemMessage(Colorize(NORMAL_COLOR, tostring(text), "DEFAULT")) end, text)
	end
	cprintf = function(text, ...)
		pcall(function(text, ...) CHAT_ROUTER:AddSystemMessage(Colorize(NORMAL_COLOR, text:format(...), "DEFAULT")) end, text, ...)
	end
else
	cprint = d
	cprintf = df
end

local function SplitCommandLineArgument(str, maxArg)
-- This function splits the command line arguments and returns the following lua table.
-- maxArg is optional and limits the maximum number of arguments that can be stored. If not specified, all arguments will be stored.
--
-- (1) the arguments separated by spaces are divided and stored in the numeric table.
--    (example) lua string "arg1 arg2 -option=x"  --->  luaTable[1] = "arg1", luaTable[2] = "arg2", luaTable[3] = "-option=x"
--
-- (2) Then, the option specifiers are analyzed and stored in the associative array part of the same lua table.
--     If each argument starts with a minus or a slash, it is considered as an option specifier.
--    (example)  -("option1")=("value")   --->  luaTable["option1"] = "value"
--               /("option2")             --->  luaTable["option2"] = true
--
-- NOTE : when referencing each split argument in a loop, you need to use only the numeric part using ipairs() instead of pairs().
--
	if type(str) == "string" then
		local optionList = {}
		local n = 0
--		local check = function(v) optionList[#optionList + 1] = v end
		local check = function(v)
			local opt, value = v:match("^[-/]([^=]+)%=*(.*)")
			optionList[#optionList + 1] = v
			if opt then
				if value == "" then value = true end
				optionList[opt] = value
			end
		end
		
		str, n = str:gsub("(%S+)", check, maxArg)
		if n > 0 then
			return optionList
		end
	end
end

-- ---------------------------------------------------------------------------------------
local db_default = {
}
local sv_default = {
}
function CDIAG:OnAddOnLoaded()
	self.lang = GetCVar("Language.2")
	self.isGamepad = IsInGamepadPreferredMode()
	self.currentApiVersion = GetAPIVersion()
	self.isFirstTimePlayerActivated = true
	self.isMasterListCreated = false
	self.requestRebuildMasterList = false
	self.currentLoadOutOfDateAddons = am:GetLoadOutOfDateAddOns()	-- LoadOutOfDateAddons flag when creating master list
	self.masterList = {}
	self.referencedList = {}
	self.lutIndexToName = {}
	self.lutNameToIndex = {}
	self.forceDisabledList = {}

	self.db = ZO_SavedVars:NewAccountWide(self.savedVarsDB, 1, nil, db_default)
	if self.db.apiVersion == nil then self.db.apiVersion = self.currentApiVersion end
	if self.db.loadOutOfDateAddons == nil then self.db.loadOutOfDateAddons = am:GetLoadOutOfDateAddOns() end
	if self.db.shouldWarnOutOfDateAddons == nil then self.db.shouldWarnOutOfDateAddons = am:ShouldWarnOutOfDateAddOns() end
--	self.svCharacter = ZO_SavedVars:NewCharacterIdSettings(self.savedVarsSV, 1, nil, sv_default, GetWorldName())

	ZO_PreHook(am, "SetAddOnEnabled", function(ADDON_MANAGER_self, addOnIndex, enabled)
		self.LDL:Debug("SetAddOnEnabled : index=%s, enabled=%s", tostring(addOnIndex), tostring(enabled))
		self.requestRebuildMasterList = true
	end)
	EVENT_MANAGER:RegisterForEvent(self.name, EVENT_FORCE_DISABLED_ADDONS_UPDATED, function(event)
		self.LDL:Debug("ForceDisabledAddOnsUpdated : num=%s", tostring(am:GetNumForceDisabledAddOns()))
		self.requestRebuildMasterList = true
	end)
	EVENT_MANAGER:RegisterForEvent(self.name, EVENT_PLAYER_DEACTIVATED, function(event)
		self:OnPlayerDeactivated(event)
	end)
	EVENT_MANAGER:RegisterForEvent(self.name, EVENT_PLAYER_ACTIVATED, function(event, initial)
		self:OnPlayerActivated(event, initial)
	end)
	self:RegisterMainCommand()
	self:RegisterKeybindsCommand()
	self:RegisterResetCommand()
end

function CDIAG:OnPlayerDeactivated(event)
	self.LDL:Debug("EVENT_PLAYER_DEACTIVATED :")
	self.db.apiVersion = self.currentApiVersion
	self.db.loadOutOfDateAddons = am:GetLoadOutOfDateAddOns()
	self.db.shouldWarnOutOfDateAddons = am:ShouldWarnOutOfDateAddOns()
end

function CDIAG:OnPlayerActivated(event, initial)
	self.LDL:Debug("EVENT_PLAYER_ACTIVATED : initial =", initial, ", isFirstTime =", self.isFirstTimePlayerActivated)
	if self.isFirstTimePlayerActivated then
		self.isFirstTimePlayerActivated = false
		if initial then	-- --------------------------------- after login
			-- Check if this is the first login after api version update since using this addon
			if self.db.apiVersion < self.currentApiVersion then
				self.LDL:Debug("detected version up")
				if am:GetLoadOutOfDateAddOns() == false and self.db.loadOutOfDateAddons == true then
					self:AdvisoryServiceOutOfDateAddonsFlag()
				end
				if am:ShouldWarnOutOfDateAddOns() == true and CDIAG.db.shouldWarnOutOfDateAddons == false then
					self:AdvisoryServiceShouldWarnOutOfDateAddonsFlag()
				end
			end
		else	-- ----------------------------------------- after reloadui
			if am:ShouldWarnOutOfDateAddOns() == true then
				self:RegisterReloadCommand()
			end
		end
	end
end

function CDIAG:BuildAddonMasterList()
	if self.isMasterListCreated then return end
	local addonList = {}
	local libList = {}
	for i = 1, am:GetNumAddOns() do
		local name, title, author, description, enabled, state, isOutOfDate, isLibrary = am:GetAddOnInfo(i)
        local entryData = {
			addonIndex = i, 
			name = Decolorize(name), 
			title = Decolorize(title), 
			author = Decolorize(author), 
			description = Decolorize(description), 
			enabled = enabled, 
			state = state,
			isOutOfDate = isOutOfDate, 
			isLibrary = isLibrary, 
			path = am:GetAddOnRootDirectoryPath(i):gsub("^user:/", ""):gsub("/$", ""), 
			version = am:GetAddOnVersion(i), 
		}

		-- investigating tier1 dependencies
		local numDependencies = am:GetAddOnNumDependencies(i)
		if numDependencies > 0 then
			local dependencies = {}
			for j = 1, numDependencies do
				local dependencyName, exists, active, minVersion, loadedVersion = am:GetAddOnDependencyInfo(i, j)
				local dependencyEntryData = {
					name = dependencyName, 
					exists = exists, 
					active = active, 
					minVersion = minVersion, 
					loadedVersion = loadedVersion, 
					error = not (exists and active and (loadedVersion >= minVersion)), 
				}
				dependencies[dependencyName] = dependencyEntryData
			end
			entryData.dependencies = dependencies
		end
		self.masterList[name] = entryData

		if isLibrary == true then
			libList[#libList + 1] = name
		else
			addonList[#addonList + 1] = name
		end
	end
	table.sort(addonList, function(a,b) return zo_strlower(self.masterList[a].title) < zo_strlower(self.masterList[b].title) end)	-- alphabetical sort by title
	table.sort(libList, function(a,b) return zo_strlower(self.masterList[a].title) < zo_strlower(self.masterList[b].title) end)	-- alphabetical sort by title

	-- making LookUpTables
	for i, v in pairs(addonList) do
		self.lutIndexToName[i] = v
		self.lutNameToIndex[v] = i
	end
	for i, v in pairs(libList) do
		self.lutIndexToName[i + #addonList] = v
		self.lutNameToIndex[v] = i + #addonList
	end

	-- making referencedList
	for k, v in pairs(self.masterList) do
		if v.dependencies then
			for referencedName, _ in pairs(v.dependencies) do
				local entries = self.referencedList[referencedName] or {}
				entries[#entries + 1] = k
				self.referencedList[referencedName] = entries
			end
		end
	end

	-- making forceDisabledList
	for i = 1, am:GetNumForceDisabledAddOns() do
		local disabledAddonName, shouldShowNotification = am:GetForceDisabledAddOnInfo(i)
		self.forceDisabledList[i] = disabledAddonName
	end

	self.currentLoadOutOfDateAddons = am:GetLoadOutOfDateAddOns()
	self.isMasterListCreated = true
	self.requestRebuildMasterList = false
	self.LDL:Debug("Add-on master list created")
end

function CDIAG:ConfirmMasterList()
	if not self.isMasterListCreated then
		self:BuildAddonMasterList()
	end
	if self.requestRebuildMasterList or am:GetLoadOutOfDateAddOns() ~= self.currentLoadOutOfDateAddons then
		-- Need to rebuild master list
		ZO_ClearTable(self.masterList)
		ZO_ClearTable(self.referencedList)
		ZO_ClearTable(self.lutIndexToName)
		ZO_ClearTable(self.lutNameToIndex)
		ZO_ClearTable(self.forceDisabledList)
		self.isMasterListCreated = false
		self:BuildAddonMasterList()
	end
end


function CDIAG:InvestigateNestedDependencies(targetAddonName)
--	Investigate totalDependencies of target add-on, using depth-first search of dependency tree with pre-order traversal
	local depth = 0
	local totalDependencies = {}
	local verifiedNode = {}
	local appearedNode = {}
	local circularDependencies
	-- circularDependencies is a numeric table that contains 'source' and 'destination' when a circular dependency is found.

	local function TraverseDependencyGraph(addonName)
		depth = depth + 1
--		self.LDL:Debug("Node[%d] : %s", depth, addonName)
		local dependencyEdge = self.masterList[addonName].dependencies
		appearedNode[addonName] = true
		if dependencyEdge then	-- not leaf
			for edge, v in pairs(dependencyEdge) do
				if appearedNode[edge] then	-- checking a circular dependency to avoid infinite loop
					circularDependencies = circularDependencies or {}
					circularDependencies[#circularDependencies + 1] = { source = addonName, destination = edge }
				else
					if totalDependencies[edge] then
						-- if matching the registered file, it should update the minimum version requirement if necessary
						if totalDependencies[edge].minVersion < v.minVersion then
							totalDependencies[edge].minVersion = v.minVersion
						end
					else
						totalDependencies[edge] = v
					end
					-- determines the need to investigate nested dependencies.
					if not verifiedNode[edge] and self.masterList[edge] then
						TraverseDependencyGraph(edge)
					end
				end
			end
		end
		verifiedNode[addonName] = true
		appearedNode[addonName] = nil
		depth = depth - 1
	end

	self:ConfirmMasterList()
	if type(targetAddonName) ~= "string" then return end
	if not self.masterList[targetAddonName] then return end

	TraverseDependencyGraph(targetAddonName)

	self.LDL:Debug("totalDependencies = %d", GetTableElements(totalDependencies))
	if GetTableElements(totalDependencies) > 0 then
		return totalDependencies, circularDependencies
	end
end


function CDIAG:GetAddonStatus(addonName)
	local entry = self.masterList[addonName]
	local status
	if entry then
		if entry.enabled then
			if entry.state == ADDON_STATE_ENABLED then
				if entry.isOutOfDate then
					status = "OUTDATED"
				else
					status = "ENABLED"
				end
			elseif entry.state == ADDON_STATE_VERSION_MISMATCH then
				status = "LOCKED"
			else
				status = "ERROR"
			end
		else
			status = "DISABLED"
		end
	else
		status = "MISSING"
	end
	return status
end


function CDIAG:RegisterReloadCommand()
	if type(SLASH_COMMANDS["/cdiag.reload"]) == "function" then return end

	SLASH_COMMANDS["/cdiag.reload"] = function()
		self.LDL:Debug("/cdiag.reload command!")
		am:ClearWarnOutOfDateAddOns()
        if ADD_ON_MANAGER then
			if ADD_ON_MANAGER.RefreshKeybinds then
				ADD_ON_MANAGER:RefreshKeybinds()
			end
			if ADD_ON_MANAGER.RefreshData then
				ADD_ON_MANAGER:RefreshData()
			end
		end
		ReloadUI("ingame")
	end
end

function CDIAG:AdvisoryServiceOutOfDateAddonsFlag()
	cprintf("%s %s", Colorize(LIME, "[CAddonDiagnosis]"), Colorize(DARK_ORANGE, "Advisory Service "))
	cprintf("The game API version has been updated to %s", Colorize(YELLOW, tostring(self.currentApiVersion)))
	cprintf("As a result, 'Allow out of date add-ons' flag has been reset to disabled.")
	cprintf("Some add-ons may now be suspended its operation.")
	cprintf("To turn the flag back on, please manually re-enable it in the Add-ons Manager.")
	cprintf(" ")
end

function CDIAG:AdvisoryServiceShouldWarnOutOfDateAddonsFlag()
	cprintf("%s %s", Colorize(LIME, "[CAddonDiagnosis]"), Colorize(DARK_ORANGE, "Advisory Service "))
	cprintf("The game API version has been updated to %s", Colorize(YELLOW, tostring(self.currentApiVersion)))
	cprintf("As a result, 'Should warn out of date add-ons' flag has been reset to enabled.")
	cprintf("To turn the flag back off, please execute the command %s", Colorize(CORNFLOWER_BLUE, "/cdiag.reload"))
    cprintf("or read and check the relevant notifications to disable it manually.")
	cprintf(" ")
	
	self:RegisterReloadCommand()
end


function CDIAG:DoOutputHelpMessage()
	cprint("(usage) /cdiag [<filename> || <number> || -list || -all]")
	cprint(" <filename> : diagnose the operating environment of the specified add-on")
	cprint(" <number>   : same as above, but specify with the listed index number.")
	cprint(" -list      : list add-ons you have enabled")
	cprint(" -all       : list all add-ons installed (including currently disabled)")
	cprint(" -force     : list force disabled add-ons by the system.")
end


-- valid assortType for function CDIAG:DoOutputAddonsList
local ASSORT_FUNC = {
	["LIB"] = function(v) return v.isLibrary end, 
	["ADDON"] = function(v) return not v.isLibrary end, 
	["ENABLED"] = function(v) return v.enabled end, 
	["DISABLED"] = function(v) return not v.enabled end, 
	["ENABLED_ADDON"] = function(v) return v.enabled and not v.isLibrary end, 
	["DISABLED_ADDON"] = function(v) return not v.enabled and not v.isLibrary end, 
	["ENABLED_LIB"] = function(v) return v.enabled and v.isLibrary end, 
	["DISABLED_LIB"] = function(v) return not v.enabled and v.isLibrary end, 
}
function CDIAG:DoOutputAddonsList(assortType)
	self:ConfirmMasterList()
	local assort = ASSORT_FUNC[assortType] or function() return true end
	
	for index, name in pairs(self.lutIndexToName) do
		local v = self.masterList[name]
		if assort(v) then
			local status = self:GetAddonStatus(name)
			cprintf(" No.%3d : %s%s (Ver.%d) [%s] : %s", index, ICON_TEXTURE[status] or "", v.title, v.version, v.state, v.path)
		end
	end
end

function CDIAG:DoOutputForceDisabledAddonsList()
	self:ConfirmMasterList()
	if #self.forceDisabledList > 0 then	
		for _, name in ipairs(self.forceDisabledList) do
			local index = self.lutNameToIndex[name] or 0
			local v = self.masterList[name] or {}
			local status = self:GetAddonStatus(name)
			cprintf(" No.%3d : %s%s (Ver.%d) [%s] : %s", index, ICON_TEXTURE[status] or "", name, v.version or 0, v.state or "FORCE-DISABLED", v.path or "")
		end
	else
		cprintf("no force-disabled add-ons.")
	end
end

function CDIAG:DoOutputAddonInfo(addonName)
	if type(addonName) ~= "string" then return end
	local entry = self.masterList[addonName]
	local index = self.lutNameToIndex[addonName] or 0
	local status = self:GetAddonStatus(addonName) or ""
	if entry then
		cprintf("%s%s (%s Ver.%d)", ICON_TEXTURE[status] or "", entry.title, entry.name, entry.version)
		if entry.description and entry.description ~= "" then cprintf("description: %s", entry.description) end
		cprintf("file location: %s", entry.path)
	end
end

function CDIAG:DoOutputDependencies(dependencies)
	local displayOrder = {}
	local n = 0
	local indexString
	local loadingInfo
	local reqVersionInfo
	local status

	dependencies = dependencies or {}
	if type(dependencies) == "table" then
		n = GetTableElements(dependencies)
		if n > 0 then
--			for k, v in pairs(dependencies) do self.LDL:Debug("%s", tostring(k)) end
			-- sorting
			for k, _ in pairs(dependencies) do
				displayOrder[#displayOrder + 1] = k
			end
--			for k, v in pairs(displayOrder) do self.LDL:Debug("%s = %s", tostring(k), tostring(v)) end
			table.sort(displayOrder, function(a,b) return zo_strlower(a) < zo_strlower(b) end)	-- alphabetical sort by dependency name
--			for k, v in pairs(displayOrder) do self.LDL:Debug("%s = %s", tostring(k), tostring(v)) end

			for i = 1, #displayOrder do
--				self.LDL:Debug("(%d) No.%d: %s ", i, displayOrder[i], tostring(self.lutIndexToName[displayOrder[i]]))
				local entry = dependencies[displayOrder[i]]
				if self.masterList[entry.name] then
					indexString = string.format("%3d", tonumber(self.lutNameToIndex[entry.name]))
					loadingInfo = string.format("using Ver.%d (%s)", entry.loadedVersion, self.masterList[entry.name].path)
				else
					indexString = Colorize(RED, "???")
					loadingInfo = Colorize(RED, "missing !!")
				end

				if entry.minVersion ~= 0 then
					reqVersionInfo = string.format("Ver.%d or later ", entry.minVersion)
				else
					reqVersionInfo = ""
				end

				if entry.exists then
					status = self:GetAddonStatus(entry.name)
					if entry.loadedVersion < entry.minVersion then
						status = "ERROR_VERSION"
					end
				else
					status = "MISSING"
				end

--				cprintf("%s [%s][%s] : minVer=%d, Ver=%d", entry.name, tostring(entry.exists), tostring(entry.active), entry.minVersion, entry.loadedVersion)	-- for debug
				cprintf("%sNo.%s : %s  %s-> %s", ICON_TEXTURE[status] or "・", indexString, entry.name, reqVersionInfo, loadingInfo)
			end
		else
			cprintf("no dependencies.")
		end
	end
end

function CDIAG:DoDiagnoseAddon(addonName)
	self:ConfirmMasterList()
	local totalDependencies, circularDependencies = self:InvestigateNestedDependencies(addonName)

--	for k, v in pairs(totalDependencies) do self.LDL:Debug("totalDependencies : %s", k) end

	cprintf("Diagnose add-on [%s] ...", addonName)
	if self.masterList[addonName] then
		self:DoOutputAddonInfo(addonName)
		cprint(Colorize(DARK_ORANGE, "Needed libraries/dependencies :"))
		self:DoOutputDependencies(totalDependencies)
	else
		cprintf("Error: The add-on [%s] could not found.", addonName)
	end
end


-- -------------------------------------------------------------------------------------------------

function CDIAG:RegisterDebugCommand()
	if type(SLASH_COMMANDS["/cdiag.debug"]) == "function" then return end
	SLASH_COMMANDS["/cdiag.debug"] = function(arg) if arg ~= "" then self:ConfigDebug({tonumber(arg)}) end end
end

function CDIAG:RegisterTestCommand()
	if type(SLASH_COMMANDS["/cdiag.test"]) == "function" then return end
	SLASH_COMMANDS["/cdiag.test"] = function(arg)
		local optionList = SplitCommandLineArgument(arg) or {}

--[[
	-- internal keybinds test
		if optionList["showkeybinds"] then
			local ratainedTestCases = 0
			for i, data in ipairs(keybindsSlotTestKeyboard) do
				local layer, category, action = GetActionIndicesFromName(data[1])
				if layer and category and action then
					local keyResult, mod1Result, mod2Result, mod3Result, mod4Result = GetActionBindingInfo(layer, category, action, 1)
					if keyResult == data[2] and mod1Result == (data[3] or KEY_INVALID) and mod2Result == (data[4] or KEY_INVALID) and mod3Result == (data[5] or KEY_INVALID) and mod4Result == (data[6] or KEY_INVALID) then
						ratainedTestCases = ratainedTestCases + 1
					end
				end
			end
			cprintf("retained test cases of keybindings: %s", tostring(ratainedTestCases))
		end

		if optionList["setkeybinds"] then
			local numTestCase = tonumber(optionList["setkeybinds"]) or 100
			if numTestCase > 0 and numTestCase <= #keybindsSlotTestKeyboard then
				cprintf("set keyboard bindings for test purpose : %s", tostring(numTestCase))
				cprint("To disable this key bindings, please execute /cdiag.reset command.")
				for i = 1, numTestCase do
					local layer, category, action = GetActionIndicesFromName(keybindsSlotTestKeyboard[i][1])
					if layer and category and action then
						if IsProtectedFunction("UnbindAllKeysFromAction") then
							CallSecureProtected("UnbindAllKeysFromAction", layer, category, action)
						else
							UnbindAllKeysFromAction(layer, category, action)
						end
						if IsProtectedFunction("BindKeyToAction") then
							CallSecureProtected("BindKeyToAction", layer, category, action, 1, keybindsSlotTestKeyboard[i][2], keybindsSlotTestKeyboard[i][3], keybindsSlotTestKeyboard[i][4], keybindsSlotTestKeyboard[i][5], keybindsSlotTestKeyboard[i][6])
						else
							BindKeyToAction(layer, category, action, 1, keybindsSlotTestKeyboard[i][2], keybindsSlotTestKeyboard[i][3], keybindsSlotTestKeyboard[i][4], keybindsSlotTestKeyboard[i][5], keybindsSlotTestKeyboard[i][6])
						end
					end
				end
			end
		end
]]

		if #optionList == 0 then
		-- classic test
			-- output datas
			self:ConfirmMasterList()
			if self.masterList then self.db.masterList = self.masterList end
			if self.referencedList then self.db.referencedList = self.referencedList end

			-- test icon textures
			for k, v in pairs(ICON_TEXTURE) do
				cprintf("%s : %s", k, v)
			end

		-- test function Colorize
			local sample = "TEST|CdcdcdcWHITE|Cff0000RED|RDEFAULT COLOR|R"
--			local sample = "TEST|cdcdcdcWHITE|cff0000RED|rDEFAULT COLOR|r"
			cprint(sample)
			cprint(Colorize(CORNFLOWER_BLUE, sample))
			cprint(Colorize(CORNFLOWER_BLUE, sample, "FILL"))
			cprint(Colorize(CORNFLOWER_BLUE, sample, "DEFAULT"))
			cprint(Colorize("#00ff00s", sample))
			cprint(Colorize("#00ff00s", sample, "FILL"))
			cprint(Colorize("#00ff00s", sample, "DEFAULT"))
			cprint(Colorize(nil, sample, "DEFAULT"))
			cprint(Colorize("#dummy00ff00", sample, "DEFAULT"))

		-- test advisory message	
			self:AdvisoryServiceOutOfDateAddonsFlag()
			self:AdvisoryServiceShouldWarnOutOfDateAddonsFlag()

			self.LDL:Verbose("hoge")
			self.LDL:Debug("hoge")
			self.LDL:Info("hoge")
			self.LDL:Warn("hoge")
			self.LDL:Error("hoge")
			return	-- command aborted
		end
	end
end

function CDIAG:RegisterMainCommand()
	if type(SLASH_COMMANDS["/cdiag"]) == "function" then return end
	SLASH_COMMANDS["/cdiag"] = function(arg)
		local optionList = SplitCommandLineArgument(arg)
		self:ConfirmMasterList()

--[[
		if optionList then
			self.LDL:Debug("#optionList  = %d", #optionList)
			self.LDL:Debug("in pairs trial")
			for k, v in pairs(optionList) do
				self.LDL:Debug("%s = %s", tostring(k), tostring(v))
			end
			self.LDL:Debug("in ipairs trial")
			for i, v in ipairs(optionList) do
				self.LDL:Debug("%s = %s", tostring(i), tostring(v))
			end
			self.LDL:Debug("trial end...")
		end
]]

--
		cprint(" ")
		cprintf("%s V%s  (API:%d)", Colorize(LIME, self.name), self.version, self.currentApiVersion)
		if not optionList then
			self:DoOutputHelpMessage()
			return
		end


		if optionList["debug"] then
			self:RegisterDebugCommand()
			self:RegisterTestCommand()
			return
		end
		if optionList["list"] then
			cprint(Colorize(DARK_ORANGE, "Enabled add-ons :"))
			self:DoOutputAddonsList("ENABLED_ADDON")
			cprint(Colorize(DARK_ORANGE, "Enabled libraries :"))
			self:DoOutputAddonsList("ENABLED_LIB")
			return
		end
		if optionList["all"] then
			cprint(Colorize(DARK_ORANGE, "Installed add-ons :"))
			self:DoOutputAddonsList("ADDON")
			cprint(Colorize(DARK_ORANGE, "Installed libraries :"))
			self:DoOutputAddonsList("LIB")
			return
		end
		if optionList["force"] then
			cprint(Colorize(DARK_ORANGE, "Force-Disabled add-ons :"))
			self:DoOutputForceDisabledAddonsList()
			return
		end

		-- decode the first non-option argument from the optionList table
		local firstNonOptionArgument
		for i = 1, #optionList do	-- NOTE : iterate the number of elements in the numeric table part.
			if optionList[i]:match("^[^-/]+") then
				firstNonOptionArgument = optionList[i]
				break
			end
		end
	
		self.LDL:Debug("firstNonOptionArgument = %s", tostring(firstNonOptionArgument))

		-- <filename>
		if self.masterList[firstNonOptionArgument] then
			self:DoDiagnoseAddon(firstNonOptionArgument)
			return
		end
	
		-- <number>
		local indexNumber = tonumber(firstNonOptionArgument)
		if indexNumber and firstNonOptionArgument == string.format("%d", firstNonOptionArgument) then
			self.LDL:Debug("index number found : %d", indexNumber)
			self:DoDiagnoseAddon(self.lutIndexToName[indexNumber])
			return
		end

		-- <filename : accepts lowercase and uppercase misspellings>
		local lowercaseFirstNonOptionArgument = zo_strlower(firstNonOptionArgument)
		for k, v in pairs(self.masterList) do
			if zo_strlower(k) == lowercaseFirstNonOptionArgument then
				self:DoDiagnoseAddon(k)
				return
			end
		end

		-- error : specific add-on not found.nt
		cprintf("The specified add-on [%s] could not be found.", firstNonOptionArgument)
		cprint("Please check the add-on filename for spelling mistakes.")
		cprint("You know it is necessary to install add-on files for diagnosis.")
	end
end

function CDIAG:RegisterKeybindsCommand()
	if type(SLASH_COMMANDS["/cdiag.keybinds"]) == "function" then return end
	SLASH_COMMANDS["/cdiag.keybinds"] = function(arg)
		cprint(" ")
		cprintf("%s V%s  (API:%d)", Colorize(LIME, CDIAG.name), CDIAG.version, self.currentApiVersion)
		cprint("Diagnose custom keybinds slot ...")
		local numSavedBindings = GetNumSavedKeybindings()
		local maxNumSavedBindings = GetMaxNumSavedKeybindings()
		cprintf("remaining custom keybinds slot: %s / %s", tostring(maxNumSavedBindings - numSavedBindings), tostring(maxNumSavedBindings))
	end
end


local function UnbindAllKeysFromMyActions()
	local layer, category, action = GetActionIndicesFromName("CDIAG_KEYBINDS_SLOT_TEST1")
	if layer and category then
		local _, numActions = GetActionLayerCategoryInfo(layer, category)
		for i = 1, numActions do
			if IsProtectedFunction("UnbindAllKeysFromAction") then
				CallSecureProtected("UnbindAllKeysFromAction", layer, category, i)
			else
				UnbindAllKeysFromAction(layer, category, i)
			end
		end
	end
end
function CDIAG:RegisterResetCommand()
	if type(SLASH_COMMANDS["/cdiag.reset"]) == "function" then return end
	SLASH_COMMANDS["/cdiag.reset"] = function(arg)
		if not IsUnitInCombat("player") then
			cprint(" ")
			cprintf("%s V%s  (API:%d)", Colorize(LIME, CDIAG.name), CDIAG.version, self.currentApiVersion)
			cprint("Remove all keybindings of this add-on.")
			UnbindAllKeysFromMyActions()
		else
			cprint(Colorize(RED, "Error: You cannot execute this command while in combat. Please move to a safe place such as your homestead and retry."))
		end
	end
end

